/*
    Project: TCGINA
    Version: V1.26
    File   : kfidlg.c

    Copyright (c) 2005-2009 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <tchar.h>
#include "kfidlg.h"
#include "pwdlg.h"
#include "secmem.h"
#include "misc.h"

static void InitOptionCheckbox(HWND hwndControl, DWORD dwBit, DWORD dwMask,
  DWORD dwOptions)
{
    if (dwMask & dwBit) {
        SendMessage(hwndControl, BM_SETCHECK, (dwOptions & dwBit) ?
          BST_CHECKED : BST_UNCHECKED, 0);
    } else {
        SendMessage(hwndControl, BM_SETCHECK, BST_INDETERMINATE, 0);
    }
}

static void ReadOptionCheckbox(HWND hwndControl, DWORD dwBit, DWORD *pdwMask,
  DWORD *pdwOptions)
{
    LRESULT lResult = SendMessage(hwndControl, BM_GETCHECK, 0, 0);
    if (lResult == BST_CHECKED) {
        *pdwMask    |= dwBit;
        *pdwOptions |= dwBit;
    } else if (lResult == BST_UNCHECKED) {
        *pdwMask    |= dwBit;
        *pdwOptions &= ~dwBit;
    } else {
        *pdwMask    &= ~dwBit;
        *pdwOptions &= ~dwBit;
    }
}

INT_PTR CALLBACK KFI_DialogProc(HWND hwndDlg, UINT uMsg,
  WPARAM wParam, LPARAM lParam)
{
    PKFI_DIALOG_DATA pkdd;

    switch (uMsg) {
    case WM_INITDIALOG: {
        TCHAR szTmp[20];

        pkdd = (PKFI_DIALOG_DATA) lParam;
        pkdd->hwndDescription = GetDlgItem(hwndDlg, IDC_DESCRIPTION);
        pkdd->hwndId = GetDlgItem(hwndDlg, IDC_IDENTIFIER);
        pkdd->hwndItemPassword = GetDlgItem(hwndDlg,
          IDC_ITEM_PASSWORD_EDIT);
        pkdd->hwndVolumePassword = GetDlgItem(hwndDlg,
          IDC_VOLUME_PASSWORD_EDIT);
        pkdd->hwndKey = GetDlgItem(hwndDlg,
          IDC_KEYFILE_DATA_EDIT);
        pkdd->hwndHost = GetDlgItem(hwndDlg, IDC_HOST_EDIT);
        pkdd->hwndDrive = GetDlgItem(hwndDlg, IDC_DRIVE);
        pkdd->hwndCompletePassword = GetDlgItem(hwndDlg,
          IDC_COMPLETE_PASSWORD);
        pkdd->hwndPasswordCache = GetDlgItem(hwndDlg, IDC_PASSWORD_CACHE);
        pkdd->hwndMountReadOnly = GetDlgItem(hwndDlg, IDC_MOUNT_READONLY);
        pkdd->hwndMountRemovable = GetDlgItem(hwndDlg, IDC_MOUNT_REMOVABLE);
        pkdd->hwndMountShared = GetDlgItem(hwndDlg, IDC_MOUNT_SHARED);
        pkdd->hwndUpdateTimeStamp = GetDlgItem(hwndDlg, IDC_UPDATE_TIMESTAMP);
        pkdd->hwndOk = GetDlgItem(hwndDlg, IDOK);
        SendMessage(pkdd->hwndDescription, EM_LIMITTEXT,
          pkdd->dwDescriptionSize - 1, 0);
        SendMessage(pkdd->hwndId, EM_LIMITTEXT,
          10, 0);
        SendMessage(pkdd->hwndHost, EM_LIMITTEXT,
          pkdd->dwHostNameSize - 1, 0);
        if (pkdd->pszCaption) {
            SetWindowText(hwndDlg, pkdd->pszCaption);
        }
        SetWindowText(pkdd->hwndDescription, pkdd->pszDescription);
        _sntprintf(szTmp, countof(szTmp), TEXT("0x%08X"), (unsigned)pkdd->dwID);
        SetWindowText(pkdd->hwndId, szTmp);
        if (*(LPSTR)pkdd->pPassword) {
            SetWindowText(pkdd->hwndItemPassword, TEXT("********"));
        }
        if (*(LPSTR)pkdd->pVolumePassword) {
            SetWindowText(pkdd->hwndVolumePassword, TEXT("********"));
        }
        pkdd->wTempKeySize = pkdd->wKeySize;
        if (pkdd->wKeySize) {
            if (pkdd->wKeySize == 1) {
                CopyString(szTmp, countof(szTmp), TEXT("(1 byte)"));
            } else {
                _sntprintf(szTmp, countof(szTmp), TEXT("(%u bytes)"),
                  (unsigned)pkdd->wKeySize);
            }
            SetWindowText(pkdd->hwndKey, szTmp);
        }
        SetWindowText(pkdd->hwndHost, pkdd->pszHostName);
        SendMessage(pkdd->hwndDrive, CB_ADDSTRING, 0, (LPARAM)TEXT(""));
        szTmp[1] = TEXT(':');
        szTmp[2] = TEXT('\0');
        for (szTmp[0] = TEXT('A'); szTmp[0] <= TEXT('Z'); szTmp[0]++) {
            SendMessage(pkdd->hwndDrive, CB_ADDSTRING, 0, (LPARAM)szTmp);
        }
        if (pkdd->chDrive >= TEXT('A') && pkdd->chDrive <= TEXT('Z')) {
            SendMessage(pkdd->hwndDrive, CB_SETCURSEL, pkdd->chDrive -
              (TEXT('A') - 1), 0);
        } else if (pkdd->chDrive >= TEXT('a') && pkdd->chDrive <= TEXT('z')) {
            SendMessage(pkdd->hwndDrive, CB_SETCURSEL, pkdd->chDrive -
              (TEXT('a') - 1), 0);
        } else if (pkdd->chDrive <= 26) {
            SendMessage(pkdd->hwndDrive, CB_SETCURSEL, pkdd->chDrive, 0);
        }
        switch(pkdd->wPasswordType & 3) {
        case KFIPT_INCOMPLETE:
            SendMessage(pkdd->hwndCompletePassword, BM_SETCHECK,
              BST_UNCHECKED, 0);
            break;
        case KFIPT_COMPLETE:
            SendMessage(pkdd->hwndCompletePassword, BM_SETCHECK,
              BST_CHECKED, 0);
            break;
        default:
            SendMessage(pkdd->hwndCompletePassword, BM_SETCHECK,
              BST_INDETERMINATE, 0);
            break;
        }
        InitOptionCheckbox(pkdd->hwndPasswordCache, KFIMO_CACHE_PASSWORDS,
            pkdd->dwMountMask, pkdd->dwMountFlags);
        InitOptionCheckbox(pkdd->hwndMountReadOnly, KFIMO_READ_ONLY,
            pkdd->dwMountMask, pkdd->dwMountFlags);
        InitOptionCheckbox(pkdd->hwndMountRemovable, KFIMO_REMOVABLE,
            pkdd->dwMountMask, pkdd->dwMountFlags);
        InitOptionCheckbox(pkdd->hwndMountShared, KFIMO_SHARED,
            pkdd->dwMountMask, pkdd->dwMountFlags);
        InitOptionCheckbox(pkdd->hwndUpdateTimeStamp, KFIMO_UPDATE_TIMESTAMP,
            pkdd->dwMountMask, pkdd->dwMountFlags);


        SetWindowUserData(hwndDlg, pkdd);
        return TRUE; }

    case WM_COMMAND:
        switch (wParam) {
        case IDOK: {
            LRESULT lResult;
            TCHAR   szTmp[20];
            LPTSTR  pszTmp;
            DWORD   dwTmp;

            pkdd = (PKFI_DIALOG_DATA)GetWindowUserData(hwndDlg);
            GetWindowText(pkdd->hwndId, szTmp, countof(szTmp));
            dwTmp = _tcstoul(szTmp, &pszTmp, 0);
            if (pszTmp == NULL || *pszTmp != TEXT('\0') || dwTmp == 0) {
                dwTmp = (DWORD)_tcslen(szTmp);
                SendMessage(pkdd->hwndId, EM_SETSEL, 0, dwTmp);
                SetFocus(pkdd->hwndId);
                return TRUE;
            }
            pkdd->dwID = dwTmp;
            GetWindowText(pkdd->hwndDescription, pkdd->pszDescription,
                pkdd->dwDescriptionSize);
            if (pkdd->pKey != pkdd->pTempKey) {
                memcpy(pkdd->pKey, pkdd->pTempKey, pkdd->wMaxKeySize);
            }
            GetWindowText(pkdd->hwndHost, pkdd->pszHostName,
              pkdd->dwHostNameSize);
            pkdd->wPasswordType &= ~3;
            switch (SendMessage(pkdd->hwndCompletePassword, BM_GETCHECK,
              0, 0)) {
            case BST_UNCHECKED:
                pkdd->wPasswordType |= KFIPT_INCOMPLETE;
                break;
            case BST_CHECKED:
                pkdd->wPasswordType |= KFIPT_COMPLETE;
                break;
            }
            lResult = SendMessage(pkdd->hwndDrive, CB_GETCURSEL, 0, 0);
            if (lResult != CB_ERR) {
                if (lResult == 0 ) {
                    pkdd->chDrive = '\0';
                } else {
                    pkdd->chDrive = ('A'-1) + (char)lResult;
                }
            }
            pkdd->dwMountMask = 0;
            pkdd->dwMountFlags = 0;
            ReadOptionCheckbox(pkdd->hwndPasswordCache, KFIMO_CACHE_PASSWORDS,
              &pkdd->dwMountMask, &pkdd->dwMountFlags);
            ReadOptionCheckbox(pkdd->hwndMountReadOnly, KFIMO_READ_ONLY,
              &pkdd->dwMountMask, &pkdd->dwMountFlags);
            ReadOptionCheckbox(pkdd->hwndMountRemovable, KFIMO_REMOVABLE,
              &pkdd->dwMountMask, &pkdd->dwMountFlags);
            ReadOptionCheckbox(pkdd->hwndMountShared, KFIMO_SHARED,
              &pkdd->dwMountMask, &pkdd->dwMountFlags);
            ReadOptionCheckbox(pkdd->hwndUpdateTimeStamp,
              KFIMO_UPDATE_TIMESTAMP, &pkdd->dwMountMask, &pkdd->dwMountFlags);

            /* fall through */ }

        case IDCANCEL: {
            EndDialog(hwndDlg, LOWORD(wParam));
            return TRUE; }

        case IDC_ITEM_PASSWORD_BUTTON: {
            PASSWORD_DIALOG_DATA pdd;
            pkdd = (PKFI_DIALOG_DATA)GetWindowUserData(hwndDlg);
            memset(&pdd, '\0', sizeof(pdd));
            pdd.pszCaption = NULL;
            pdd.uCodePage = CP_UTF8;
            pdd.pPassword = pkdd->pPassword;
            pdd.cbPasswordSize = KFI_MAX_PASSWORD;
            pdd.pwszTempPassword = pkdd->pwszTempPassword;
            pdd.pwszConfirmation = pkdd->pwszConfirmation;
            pdd.dwMinPasswordLength = 0;
            pdd.dwMaxPasswordLength = pkdd->dwPasswordLength;
            if (DialogBoxParam(GetModuleHandle(0),
              MAKEINTRESOURCE(IDD_PASSWORDS), NULL, PasswordDialogProc,
              (LPARAM)&pdd) == IDOK) {
                SetWindowText(pkdd->hwndItemPassword,
                  *(LPSTR)pkdd->pPassword ? TEXT("********") : TEXT(""));
            }
            return TRUE; }

        case IDC_VOLUME_PASSWORD_BUTTON: {
            PASSWORD_DIALOG_DATA pdd;
            pkdd = (PKFI_DIALOG_DATA)GetWindowUserData(hwndDlg);
            memset(&pdd, '\0', sizeof(pdd));
            pdd.pszCaption = TEXT("Enter Volume Password");
            pdd.uCodePage = CP_ACP;
            pdd.pPassword = pkdd->pVolumePassword;
            pdd.cbPasswordSize = KFI_MAX_VOLUME_PASSWORD;
            pdd.pwszTempPassword = pkdd->pwszTempPassword;
            pdd.pwszConfirmation = pkdd->pwszConfirmation;
            pdd.dwMinPasswordLength = 0;
            pdd.dwMaxPasswordLength = pkdd->dwVolumePasswordLength;
            if (DialogBoxParam(GetModuleHandle(0),
              MAKEINTRESOURCE(IDD_PASSWORDS), NULL, PasswordDialogProc,
              (LPARAM)&pdd) == IDOK) {
                SetWindowText(pkdd->hwndVolumePassword,
                  *(LPSTR)pkdd->pVolumePassword ? TEXT("********") : TEXT(""));
            }
            return TRUE; }

        case IDC_KEYFILE_DATA_BUTTON: {
            MessageBox(hwndDlg,
              TEXT("Import of keyfile data is not yet implemented!"),
              TEXT("Keyfile Editor"), MB_ICONSTOP|MB_OK);
            return TRUE; }

        case IDC_HOST_BUTTON: {
            OPENFILENAME ofn;
            pkdd = (PKFI_DIALOG_DATA)GetWindowUserData(hwndDlg);
            GetWindowText(pkdd->hwndHost, pkdd->pszTempHostName,
              pkdd->dwHostNameSize);
            memset(&ofn, '\0', sizeof(ofn));
            ofn.lStructSize = sizeof(ofn);
            ofn.hwndOwner = hwndDlg;
            ofn.lpstrFilter = TEXT("All Files\0*.*\0\0");
            ofn.lpstrFile = pkdd->pszTempHostName;
            ofn.nMaxFile = pkdd->dwHostNameSize;
            ofn.lpstrTitle = TEXT("Select Host File Name");
            ofn.Flags = OFN_HIDEREADONLY | OFN_DONTADDTORECENT;
            if (GetOpenFileName(&ofn)) {
                SendMessage(pkdd->hwndHost, WM_SETTEXT, 0,
                  (LPARAM)pkdd->pszTempHostName);
            } else if (CommDlgExtendedError() == FNERR_INVALIDFILENAME) {
                *pkdd->pszHostName = TEXT('\0');
                if (GetOpenFileName(&ofn)) {
                    SendMessage(pkdd->hwndHost, WM_SETTEXT, 0,
                        (LPARAM)pkdd->pszTempHostName);
                }
            }
            return TRUE; }
        }
        break;
    }

    return FALSE;
}

BOOL KFI_Edit(PKFI_DATA pd, void *pPassword, DWORD cbPasswordSize,
  LPCTSTR pszCaption)
{
    BOOL fSuccess;
    KFI_DIALOG_DATA pkd;
    struct {
        TCHAR szDescription[KFI_MAX_DESCRIPTION];
        BYTE  rgbPassword[136];
        BYTE  rgbVolumePassword[136];
        WCHAR wszTempPassword[136];
        WCHAR wszConfirmation[136];
        TCHAR szHostName[MAX_PATH];
        BYTE  rgbKey[1024];
    } *tmp;

    if (cbPasswordSize > sizeof(tmp->rgbPassword)) {
        return FALSE;
    }

    if (!sec_alloc(tmp)) {
        return FALSE;
    }
    memset(tmp, '\0', sizeof(*tmp));
    memset(&pkd, '\0', sizeof(pkd));
    pkd.pszCaption                  = pszCaption;
    pkd.pszDescription              = tmp->szDescription;
    pkd.dwID                        = pd->dwID;
    pkd.dwDescriptionSize           = KFI_MAX_DESCRIPTION;
    pkd.pPassword                   = tmp->rgbPassword;
    pkd.pwszTempPassword            = tmp->wszTempPassword;
    pkd.pwszConfirmation            = tmp->wszConfirmation;
    pkd.dwPasswordLength            = KFI_MAX_PASSWORD-1;
    pkd.pVolumePassword             = tmp->rgbVolumePassword;
    pkd.dwVolumePasswordLength      = KFI_MAX_VOLUME_PASSWORD-1;
    pkd.pszHostName                 = tmp->szHostName;
    pkd.pszTempHostName             = tmp->szHostName;
    pkd.dwHostNameSize              = MAX_PATH;
    pkd.chDrive                     = pd->chDriveLetter;
    pkd.wPasswordType               = pd->wPasswordType;
    pkd.dwMountMask                 = pd->dwMountMask;
    pkd.dwMountFlags                = pd->dwMountOptions;
    pkd.wKeySize                    = pd->wKeySize;
    pkd.pKey                        = tmp->rgbKey;
    pkd.pTempKey                    = tmp->rgbKey;
    pkd.wMaxKeySize                 = sizeof(pd->rgbKey);
    memcpy(tmp->rgbKey, pd->rgbKey, sizeof(pd->rgbKey));
    memcpy(tmp->rgbPassword, pPassword, cbPasswordSize);
    memcpy(tmp->rgbVolumePassword, pd->aszPassword, KFI_MAX_VOLUME_PASSWORD);
    CopyStringFromUnicode(tmp->szDescription, countof(tmp->szDescription),
      pd->wszDescription);
    CopyStringFromUnicode(tmp->szHostName, countof(tmp->szHostName),
      pd->wszHostName);

    fSuccess = DialogBoxParam(GetModuleHandle(0),
      MAKEINTRESOURCE(IDD_KEYFILE_ITEM), NULL, KFI_DialogProc,
      (LPARAM)&pkd) == IDOK;
    if (fSuccess) {
        pd->dwID           = pkd.dwID;
        pd->chDriveLetter  = pkd.chDrive;
        pd->dwMountMask    = pkd.dwMountMask;
        pd->dwMountOptions = pkd.dwMountFlags;
        pd->wPasswordType  = pkd.wPasswordType;
#if 0
        /* NOT YET IMPLEMENTED */
        pd->wKeySize       = pkd.wKeySize;
        memcpy(pd->rgbKey, tmp->rgbKey, sizeof(pd->rgbKey));
#endif
        memcpy(pPassword, tmp->rgbPassword, cbPasswordSize);
        memcpy(pd->aszPassword, tmp->rgbVolumePassword,
          KFI_MAX_VOLUME_PASSWORD);
        pd->wPasswordSize = 0;
        CopyStringToUnicode(pd->wszDescription, countof(pd->wszDescription),
          tmp->szDescription);
        CopyStringToUnicode(pd->wszHostName, countof(pd->wszHostName),
          tmp->szHostName);
    }

    sec_free(tmp);

    return fSuccess;
}
